from dataclasses import dataclass, field
from typing import Dict, Type, Any
from beet import Context
from beet.core.utils import required_field
from bolt import Accumulator, Runtime
from bolt.utils import internal
from mecha import AstNode, Mecha, Parser, Reducer, rule
from tokenstream import TokenStream, set_location
from nbtlib import Numeric, Byte, Short, Long, Float, Double
import operator

SUFFIXED_NUMBER = r"[+-]?(?:[0-9]*?\.[0-9]+|[0-9]+\.[0-9]*?|[1-9][0-9]*|0)(?:[eE][+-]?[0-9]+)?[bslfdBSLFD]\b"


# nbtlib monkey patch :)
def patch_nbtlib():
    def op_method(op: Any, reverse: Any = False):
        def method(self: Any, other: Any):
            value = op(other.real, self.real) if reverse else op(self.real, other.real)
            return self.__class__(value)

        return method

    Numeric.__add__ = op_method(operator.add)
    Numeric.__radd__ = op_method(operator.add, reverse=True)
    Numeric.__sub__ = op_method(operator.sub)
    Numeric.__rsub__ = op_method(operator.sub, reverse=True)
    Numeric.__mul__ = op_method(operator.mul)
    Numeric.__rmul__ = op_method(operator.mul, reverse=True)
    Numeric.__truediv__ = op_method(operator.truediv)
    Numeric.__rtruediv__ = op_method(operator.truediv, reverse=True)


def beet_default(ctx: Context):
    ctx.inject(nbt_literals)


def nbt_literals(ctx: Context):
    patch_nbtlib()

    mc = ctx.inject(Mecha)
    runtime = ctx.inject(Runtime)

    runtime.globals.update(
        {"Byte": Byte, "Short": Short, "Long": Long, "Float": Float, "Double": Double}
    )

    mc.spec.parsers["bolt:literal"] = NbtLiteralParser(
        parser=mc.spec.parsers["bolt:literal"]
    )

    runtime.modules.codegen.extend(NbtLiteralCodegen())

    runtime.helpers["nbt_literal"] = NbtLiteralConverter()


@dataclass(frozen=True)
class AstNbtLiteral(AstNode):
    value: float = required_field()
    suffix: str = required_field()


@dataclass
class NbtLiteralParser:
    parser: Parser

    def __call__(self, stream: TokenStream):
        with stream.syntax(nbt_literal=SUFFIXED_NUMBER):
            number = stream.get("nbt_literal")

            if not number:
                return self.parser(stream)

            value = float(number.value[:-1])
            suffix = number.value[-1].lower()

            node = AstNbtLiteral(value=value, suffix=suffix)

            return set_location(node, stream.current)


class NbtLiteralCodegen(Reducer):
    @rule(AstNbtLiteral)
    def nbt_literal(self, node: AstNbtLiteral, acc: Accumulator):
        result = acc.make_variable()
        rhs = acc.helper("nbt_literal", node.value, repr(node.suffix))

        acc.statement(f"{result} = {rhs}")

        return [result]


@dataclass
class NbtLiteralConverter:
    number_suffixes: Dict[str, Type[Numeric]] = field(
        default_factory=lambda: {
            "b": Byte,
            "s": Short,
            "l": Long,
            "f": Float,
            "d": Double,
        }
    )

    @internal
    def __call__(self, value: float, suffix: str):
        type = self.number_suffixes.get(suffix)

        if type is None:
            return value

        return type(value)



__version__ = "1.0.2"


from beet import Context
from beet.contrib.load import load
from bolt import bolt


def beet_default(ctx: Context):
    ctx.require(
        load(
            data_pack={"data/reapermc/modules": "@reapermc/modules"},
        ),
        bolt(entrypoint=["*:main", "reapermc:core"]),
        "bolt_expressions.plugin",
        "reapermc.plugins.nbt_literals",
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reapermc', 'reapermc.plugins']

package_data = \
{'': ['*'],
 'reapermc': ['modules/*', 'modules/event/*', 'modules/flag/*', 'resources/*']}

install_requires = \
['beet>=0.80.1', 'bolt-expressions>=0.11.0', 'bolt>=0.21.0', 'mecha>=0.61.0']

setup_kwargs = {
    'name': 'reapermc',
    'version': '1.0.2',
    'description': 'A high level framework for the Bolt scripting language.',
    'long_description': "# ReaperMC\n[![GitHub Actions](https://github.com/reapermc/reapermc/workflows/CI/badge.svg)](https://github.com/reapermc/reapermc/actions)\n\n> A high level framework for the Bolt scripting language.\n\n\n\n\n## Introduction\n\nReaper is a framework for the [Bolt](https://github.com/mcbeet/bolt) scripting language. It features complex yet optimized, high level yet simple to use functions.\n\n\nThe goal of this project is to reduce the time developers waste. The aim is to enable developers in creating things faster while focusing on the important bits.\n\nFor further info about the project, check [here](about.md).\n\n## Installation\n\n```bash\npip install reapermc\n```\n\n\n\n## Getting started\n\nThis package is designed to be used within `.bolt` scripts, inside a bolt enabled project. I will never officially support `.mcfunction` files.\n\nTo enable Reaper inside your project, you will need to add `reapermc` to your `require` list inside the beet config file.\n```yaml\nrequire:\n    - bolt\n    - reapermc\n```\n\n\n\nModules named `<namespace>:main` will automatically be the entrypoint.\n\nTo use Reaper's functions, you'll need to import them. (**This won't be necessary in the future**).\n\n```py\n# my_project:main\n\nfrom reapermc:api import sleep, set_time, tag\n\ndef become_wizard():\n    tellraw @s 'You will become a wizard...'\n\n    with sleep(40):\n        set_time(13000)\n        tag('wizardman')\n        tellraw @s 'You are now a wizard!'\n```\n\n\n# Documentation\n\nThe documentation for this project can be found [here](docs/table_of_contents.md).\n\n\n\n---\n\nLicense - [MIT](https://github.com/reapermc/reapermc/blob/main/LICENSE)\n\n",
    'author': 'Yeti',
    'author_email': 'arcticyeti1@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/reapermc/reapermc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
